from ._condition import ConditionTree, Condition
from .enums import Operator
from .protocols import IEntity


class Where:
    def __init__(self, entity: IEntity | None = None, logic='and') -> None:
        self._condition_tree = ConditionTree(logic)
        self._entity: IEntity = entity

    def tree(self) -> ConditionTree:
        return self._condition_tree

    def check_field(self, field: str):
        if self._entity is not None and not hasattr(self._entity, field):
            raise ValueError(f'entity {self._entity} has no field {field}')

    def eq(self, field: str, value: any) -> 'Where':
        self.check_field(field)
        self._condition_tree.add_condition(Condition(field, value))
        return self

    def ne(self, field: str, value: any) -> 'Where':
        self.check_field(field)
        self._condition_tree.add_condition(Condition(field, value, Operator.NE))
        return self

    def gt(self, field: str, value: any) -> 'Where':
        self.check_field(field)
        self._condition_tree.add_condition(Condition(field, value, Operator.GT))
        return self

    def ge(self, field: str, value: any) -> 'Where':
        self.check_field(field)
        self._condition_tree.add_condition(Condition(field, value, Operator.GE))
        return self

    def lt(self, field: str, value: any) -> 'Where':
        self.check_field(field)
        self._condition_tree.add_condition(Condition(field, value, Operator.LT))
        return self

    def le(self, field: str, value: any) -> 'Where':
        self.check_field(field)
        self._condition_tree.add_condition(Condition(field, value, Operator.LE))
        return self

    def in_(self, field: str, value: any) -> 'Where':
        self.check_field(field)
        self._condition_tree.add_condition(Condition(field, value, Operator.IN))
        return self

    def l_like(self, field: str, value: any) -> 'Where':
        self.check_field(field)
        self._condition_tree.add_condition(Condition(field, f'%{value}', Operator.LIKE))
        return self

    def r_like(self, field: str, value: any) -> 'Where':
        self.check_field(field)
        self._condition_tree.add_condition(Condition(field, f'{value}%', Operator.LIKE))
        return self

    def like(self, field: str, value: any) -> 'Where':
        self.check_field(field)
        self._condition_tree.add_condition(Condition(field, f'%{value}%', Operator.LIKE))
        return self

    def or_(self, or_conditions: 'Or') -> 'Where':
        self._condition_tree.add_tree(or_conditions._condition_tree)
        return self

    def count(self):
        return self._condition_tree.count()


class Or(Where):
    def __init__(self, entity: IEntity | None = None) -> None:
        super().__init__(entity, 'or')


def where(entity: IEntity | None = None):
    return Where(entity)


def or_(entity: IEntity | None = None):
    return Or(entity)



from typing import Any

from loguru import logger
from pymysql.connections import Connection
from pymysql.cursors import DictCursor


def debug(sql, args):
    logger.debug(f'#### sql: {sql}')
    logger.debug(f'#### args: {args}')


def select_one(sql: str, args: tuple[Any, ...], conn: Connection) -> dict[str, Any] | None:
    debug(sql, args)

    cursor: DictCursor = conn.cursor()
    try:
        sql = sql.replace('?', '%s')
        result = cursor.execute(sql, args)
        if result is None:
            return None

        row = cursor.fetchone()
        if row is None:
            return None

        return row
    except Exception as e:
        raise e
    finally:
        cursor.close()
        conn.close()


def select_many(sql: str, args: tuple[Any, ...], conn: Connection) -> tuple[dict[str, Any], ...] | None:
    debug(sql, args)

    cursor: DictCursor = conn.cursor()
    try:
        sql = sql.replace('?', '%s')
        result = cursor.execute(sql, args)
        if result is None:
            return None

        rows = cursor.fetchall()
        if rows is None:
            return None

        return rows
    except Exception as e:
        raise e
    finally:
        cursor.close()
        conn.close()


def execute(sql: str, args: tuple[Any, ...], conn: Connection, is_insert=False) -> (int, int):
    debug(sql, args)

    conn.begin()
    cursor: DictCursor = conn.cursor()
    try:
        sql = sql.replace('?', '%s')
        row_affected = cursor.execute(sql, args)
        last_row_id = 0
        if is_insert:
            last_row_id = cursor.lastrowid
        conn.commit()
        return row_affected, last_row_id
    except Exception as e:
        conn.rollback()
        raise e
    finally:
        cursor.close()
        conn.close()


def executemany(sql: str, args: list[tuple[any, ...]], conn) -> int:
    debug(sql, args)
    conn.begin()
    cursor = conn.cursor()
    try:
        sql = sql.replace('?', '%s')
        row_affected = cursor.executemany(sql, args)
        conn.commit()
        return row_affected
    except Exception as e:
        conn.rollback()
        raise e
    finally:
        cursor.close()
        conn.close()



from typing import Protocol, runtime_checkable


@runtime_checkable
class IDataSource(Protocol):
    def get_id(self) -> str:
        ...

    def get_host(self) -> str:
        ...

    def get_port(self) -> int:
        ...

    def get_user(self) -> str:
        ...

    def get_password(self):
        ...

    def get_default_database(self) -> str:
        ...

    def get_connection(self):
        ...



from typing import Any

from ._dorm import dorm
from ._models import models
from ._mysql_executor import execute
from ._where import Where, Or
from .protocols import IEntity, IDataSource


class Update:
    def __init__(self, table: str | IEntity, datasource: IDataSource, database: str):
        if isinstance(table, str):
            self._table: str = table
            self._entity: IEntity | None = None
        else:
            self._table: str = table.__table_name__
            self._entity: IEntity | None = table

        self._datasource = datasource
        self._database = database

        if self._table is None or self._table == '':
            raise ValueError('table is required')
        if self._datasource is None or self._datasource == '':
            raise ValueError('datasource is required')
        if not isinstance(self._datasource, IDataSource):
            raise ValueError('datasource must be an instance of IDataSource')
        if self._database is None or self._database == '':
            raise ValueError('database is required')

        self._where = Where(self._entity)
        self._update_fields = {}

        self._model: callable = models.get(data_source=self._datasource, database=self._database, table=self._table)

    def eq(self, field: str, value: Any) -> 'Update':
        self._where.eq(field, value)
        return self

    def ne(self, field: str, value: Any) -> 'Update':
        self._where.ne(field, value)
        return self

    def gt(self, field: str, value: Any) -> 'Update':
        self._where.gt(field, value)
        return self

    def ge(self, field: str, value: Any) -> 'Update':
        self._where.ge(field, value)
        return self

    def lt(self, field: str, value: Any) -> 'Update':
        self._where.lt(field, value)
        return self

    def le(self, field: str, value: Any) -> 'Update':
        self._where.le(field, value)
        return self

    def in_(self, field: str, value: Any) -> 'Update':
        self._where.in_(field, value)
        return self

    def l_like(self, field: str, value: Any) -> 'Update':
        self._where.l_like(field, value)
        return self

    def r_like(self, field: str, value: Any) -> 'Update':
        self._where.r_like(field, value)
        return self

    def like(self, field: str, value: Any) -> 'Update':
        self._where.like(field, value)
        return self

    def or_(self, or_: Or) -> 'Update':
        self._where.or_(or_)
        return self

    def set(self, **sets) -> 'Update':
        for k, v in sets.items():
            if self._entity is not None and not hasattr(self._entity, k):
                raise ValueError(f'entity {self._entity} has no field {k}')
        self._update_fields: dict = sets
        return self

    def update(self) -> int:
        if self._where.count() == 0:
            # danger operation
            raise ValueError('update all is not supported')

        if len(self._update_fields) == 0:
            raise ValueError('update fields is required')

        sql, args = self._build_update()
        conn = self._datasource.get_connection()
        affected, last_row_id = execute(sql, args, conn, False)
        return affected or 0

    def _build_update(self) -> tuple[str, tuple[Any, ...]]:
        sql = f'UPDATE {self._database}.{self._table} SET {",".join([f"{k}=?" for k in self._update_fields.keys()])}'
        args = tuple(self._update_fields.values())

        exp, args2 = self._where.tree().parse()
        sql += ' WHERE ' + exp
        args += args2
        return sql, args

    def delete(self) -> int:
        if self._where.count() == 0:
            # danger operation
            raise ValueError('delete all is not supported')

        sql, args = self._build_delete()
        conn = self._datasource.get_connection()
        affected, last_row_id = execute(sql, args, conn, False)
        return affected or 0

    def _build_delete(self) -> tuple[str, tuple[Any, ...]]:
        sql = f'DELETE FROM {self._database}.{self._table}'

        exp, args = self._where.tree().parse()
        sql += ' WHERE ' + exp
        return sql, args


def update(table: str | IEntity, database: str | None = None, data_source: IDataSource | None = None) -> Update:
    return Update(table, data_source or dorm.default_datasource(), database or dorm.default_datasource().get_default_database())



from dataclasses import fields
from typing import Any

from ._dorm import dorm
from ._models import models
from ._mysql_executor import execute, executemany
from .protocols import IEntity, IDataSource


class Insert:
    def __init__(self, table: str | IEntity, datasource: IDataSource, database: str):
        if isinstance(table, str):
            self._table: str = table
            self._entity: IEntity | None = None
        else:
            self._table: str = table.__table_name__
            self._entity: IEntity | None = table

        self._datasource = datasource
        self._database = database

        if self._table is None or self._table == '':
            raise ValueError('table is required')
        if self._datasource is None or self._datasource == '':
            raise ValueError('datasource is required')
        if not isinstance(self._datasource, IDataSource):
            raise ValueError('datasource must be an instance of IDataSource')
        if self._database is None or self._database == '':
            raise ValueError('database is required')

        self._model: callable = models.get(data_source=self._datasource, database=self._database, table=self._table)

    def insert(self, data: dict) -> (int, int):
        if data is None:
            raise ValueError('null data')

        sql, args = self._build_insert(data)
        conn = self._datasource.get_connection()
        return execute(sql, args, conn, True)

    def _build_insert(self, data: dict) -> tuple[str, tuple[Any, ...]]:
        placeholder = []
        keys = []
        values = []
        model_fields = [f.name for f in fields(self._model)]
        for k, v in data.items():
            if k in model_fields:
                keys.append(k)
                values.append(v)
                placeholder.append('?')

        if len(keys) == 0:
            raise ValueError('no valid field found')

        sql = f'INSERT INTO {self._database}.{self._table} ({",".join(keys)}) VALUES ({",".join(placeholder)})'
        args = tuple(values)
        return sql, args

    def insert_bulk(self, data: list[dict]) -> int:
        if data is None or len(data) == 0:
            raise ValueError('data is required')

        sql, args = self._build_insert_bulk(data)
        conn = self._datasource.get_connection()
        return executemany(sql, args, conn)

    def _build_insert_bulk(self, data: list[dict]) -> tuple[str, list[tuple[any, ...]]]:
        placeholder = []
        keys = []
        values = []
        model_fields = [f.name for f in fields(self._model)]
        for k, v in data[0].items():
            if k in model_fields:
                keys.append(k)
                values.append(v)
                placeholder.append('?')
        sql = f'INSERT INTO {self._database}.{self._table} ({",".join(keys)}) VALUES ({",".join(placeholder)})'
        args = [tuple(datum.values()) for datum in data]
        return sql, args

    def upsert(self, data, update_fields: list[str] = None) -> (int, int):
        if data is None:
            raise ValueError('null data')

        sql, args = self._build_upsert(data, update_fields)
        conn = self._datasource.get_connection()
        return execute(sql, args, conn, True)

    def _build_upsert(self, data: dict, update_fields: list[str]) -> tuple[str, tuple[Any, ...]]:
        placeholder = []
        keys = []
        values = []
        model_fields = [f.name for f in fields(self._model)]

        for k, v in data.items():
            if k in model_fields:
                keys.append(k)
                values.append(v)
                placeholder.append('?')

        if len(keys) == 0:
            raise ValueError('no valid field found')

        # check if update_fields is a sub list of keys
        if not all(field in keys for field in update_fields):
            raise ValueError('update fields are not valid')

        sql = f'INSERT INTO `{self._database}`.`{self._table}` ({",".join(keys)}) VALUES ({",".join(placeholder)})'
        if update_fields:
            sql += f' ON DUPLICATE KEY UPDATE {",".join([f"{k}=VALUES({k})" for k in update_fields])}'
        else:
            sql += f' ON DUPLICATE KEY UPDATE {",".join([f"{k}=VALUES({k})" for k in keys])}'

        args = tuple(values)
        return sql, args

    def upsert_bulk(self, data: list[dict], update_fields: list[str]) -> int:
        if data is None or len(data) == 0:
            raise ValueError('data is required')

        sql, args = self._build_upsert_bulk(data, update_fields)
        conn = self._datasource.get_connection()
        return executemany(sql, args, conn)

    def _build_upsert_bulk(self, data: list[dict], update_fields: list[str]) -> tuple[str, list[tuple[any, ...]]]:
        if not data:
            raise ValueError('data is required')

        placeholder = []
        keys = []
        model_fields = [f.name for f in fields(self._model)]

        for k in data[0].keys():
            if k in model_fields:
                keys.append(k)
                placeholder.append('%s')

        if len(keys) == 0:
            raise ValueError('no valid field found')

        # Check if update_fields is a sublist of keys
        if not all(field in keys for field in update_fields):
            raise ValueError('update_fields must be a sublist of keys')

        sql = f'INSERT INTO `{self._database}`.`{self._table}` ({",".join(keys)}) VALUES ({",".join(placeholder)})'
        if update_fields:
            sql += f' ON DUPLICATE KEY UPDATE {",".join([f"{k}=VALUES({k})" for k in update_fields])}'
        else:
            sql += f' ON DUPLICATE KEY UPDATE {",".join([f"{k}=VALUES({k})" for k in keys])}'

        args = [tuple(datum[k] for k in keys) for datum in data]
        return sql, args


def insert(table: str | IEntity, data: dict = None, database: str | None = None, data_source: IDataSource | None = None) -> (int, int):
    return Insert(table, data_source or dorm.default_datasource(), database or dorm.default_datasource().get_default_database()).insert(data)


def upsert(table: str | IEntity, data: dict = None, update_fields: list = None, database: str | None = None,
           data_source: IDataSource | None = None) -> (int, int):
    return (Insert(table, data_source or dorm.default_datasource(), database or dorm.default_datasource().get_default_database())
            .upsert(data, update_fields))


def insert_bulk(table: str | IEntity, data: list[dict] = None, database: str | None = None, data_source: IDataSource | None = None):
    return Insert(table, data_source or dorm.default_datasource(), database or dorm.default_datasource().get_default_database()).insert_bulk(data)


def upsert_bulk(table: str | IEntity, data: list[dict] = None, update_fields: list = None, database: str | None = None,
                data_source: IDataSource | None = None):
    return (Insert(table, data_source or dorm.default_datasource(), database or dorm.default_datasource().get_default_database())
            .upsert_bulk(data, update_fields))
